import math
import collections
import collections.abc
import warnings

from itertools import combinations
from operator import __mul__
from statistics import mean
from typing import Any, Sequence, Mapping, Hashable, Optional

from coba.encodings import InteractionsEncoder
from coba.exceptions import CobaExit
from coba.statistics import percentile
from coba.utilities import PackageChecker, peek_first
from coba.primitives import Environment

#this has largely been deprecated at this point...
class ClassMetaEvaluator:#pragma: no cover
    """Describe an Environment made from a Classification dataset.

    This evlaluator calculates a number of statistics which describe a given environment. These
    statistics can be used to analyze what kinds of environments a given learner performs best on.
    To make the most of this evaluator sklearn should be installed.
    """

    def evaluate(self, environment: Optional[Environment]) -> Mapping[Any,Any]:

        #sources:
        #[1]: https://arxiv.org/pdf/1808.03591.pdf (lorena2019complex)
        #[2]: https://link.springer.com/content/pdf/10.1007/978-3-540-31883-5.pdf#page=468 (castiello2005meta)
        #[3]: https://link.springer.com/content/pdf/10.1007/s10044-012-0280-z.pdf (reif2014automatic)
        #[4]: https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.440.6255&rep=rep1&type=pdf

        #[3] found that information theoretic measures and landmarking measures are most important

        interactions = peek_first(environment.read())[1]

        if not interactions: return {}

        contexts, actions, rewards = zip(*[ (i['context'], i['actions'], i['rewards']) for i in interactions ])
        env_stats = {}

        X = [ InteractionsEncoder('x').encode(x=c) for c   in contexts             ]
        Y = [ max(a,key=r)                         for a,r in zip(actions,rewards) ]
        X = self._dense(X)

        classes = list(set(Y))
        feats   = list(range(len(X[0])))

        n = len(X)
        m = len(feats)
        k = len(classes)

        X_bin = self._bin(X,n/10)
        X_bin_by_f = list(zip(*X_bin))

        entropy_Y = self._entropy(Y)
        entropy_X = [self._entropy(x) for x in X_bin_by_f]

        mutual_XY_infos  = sorted([self._mutual_info(x,Y) for x in X_bin_by_f], reverse=True)

        #Information-Theoretic Meta-features
        env_stats["class_count"          ] = k
        env_stats["class_entropy"        ] = entropy_Y                # [1]
        env_stats["class_entropy_N"      ] = self._entropy_normed(Y)  # [1,2,3]
        env_stats["class_imbalance_ratio"] = self._imbalance_ratio(Y) # [1]

        env_stats["feature_numeric_count"  ] = sum([int(len(set(X_bin_by_f[f]))!=2) for f in feats])
        env_stats["feature_onehot_count"   ] = sum([int(len(set(X_bin_by_f[f]))==2) for f in feats])
        env_stats["feature_entropy_mean"   ] = mean([self._entropy(X_bin_by_f[f]) for f in feats])
        env_stats["feature_entropy_mean_N" ] = mean([self._entropy_normed(X_bin_by_f[f]) for f in feats])
        env_stats["joint_XY_entropy_mean"  ] = mean([self._entropy(list(zip(X_bin_by_f[f],Y))) for f in feats]) #[2,3]
        env_stats["joint_XY_entropy_mean_N"] = mean([self._entropy_normed(list(zip(X_bin_by_f[f],Y))) for f in feats])
        env_stats["mutual_XY_info_mean"    ] = mean(mutual_XY_infos) #[2,3]
        env_stats["mutual_XY_info_mean_N"  ] = mean(mutual_XY_infos)/entropy_Y if entropy_Y else None #[2,3]

        env_stats["mutual_XY_info_rank1"   ] = mutual_XY_infos[0]
        env_stats["mutual_XY_info_rank2"   ] = mutual_XY_infos[1] if len(mutual_XY_infos) > 1 else None
        env_stats["equivalent_num_X_attr"  ] = entropy_Y/mean(mutual_XY_infos) if mean(mutual_XY_infos) else None #[2,3]
        env_stats["noise_signal_ratio"     ] = (mean(entropy_X)-mean(mutual_XY_infos))/mean(mutual_XY_infos) if mean(mutual_XY_infos) else None #[2]

        env_stats["max_fisher_discrim"    ] = self._max_fisher_discriminant_ratio(X, Y) #[1]
        #env_stats["max_fisher_discrim_dir"] = self._max_directional_fisher_discriminant_ratio(X, Y) #[1] (this dies on large feature)
        env_stats["volume_overlapping"    ] = self._volume_overlapping_region(X,Y) #[1]
        env_stats["max_single_feature_eff"] = self._max_individual_feature_efficiency(X,Y) #[1]

        #Sparsity/Dimensionality measures [1,2,3]
        env_stats["feature_count"       ] = m
        env_stats["percent_nonzero_feat"] = mean([len(list(filter(None,x)))/len(x) for x in X])

        try:

            PackageChecker.sklearn("ClassEnvironmentTask")

            import numpy as np

            from sklearn.decomposition import TruncatedSVD
            from sklearn.neighbors import KNeighborsClassifier
            from sklearn.naive_bayes import GaussianNB
            from sklearn.metrics import f1_score, accuracy_score
            from sklearn.model_selection import cross_validate
            from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
            from sklearn.ensemble import RandomForestClassifier
            from sklearn.tree import DecisionTreeClassifier

            import sklearn.exceptions
            warnings.filterwarnings("ignore", category=sklearn.exceptions.FitFailedWarning)

            np_X = np.array(X)
            np_Y = np.array(Y)

            try:
                #1NN OOB [3,4]
                oob = np_Y[KNeighborsClassifier(n_neighbors=1).fit(np_X,np_Y).kneighbors(np_X, n_neighbors=2, return_distance=False)[:,1]]
                env_stats["1nn_accuracy"] = float(accuracy_score(np_Y,oob))
                env_stats["1nn_f1_macro"] = float(f1_score(np_Y,oob, average='macro'))
            except: #pragma: no cover
                pass

            try:
                #LDA [3,4]
                scr = cross_validate(LinearDiscriminantAnalysis(), np_X, np_Y, scoring=('accuracy','f1_macro'))
                env_stats["lda_accuracy"] = float(mean(scr['test_accuracy']))
                env_stats["lda_f1_weighted"] = float(mean(scr['test_f1_weighted']))
            except: #pragma: no cover
                pass

            try:
                #Naive Bayes [3,4]
                scr = cross_validate(GaussianNB(), np_X, np_Y, scoring=('accuracy','f1_macro'))
                env_stats["naive_bayes_accuracy"] = float(mean(scr['test_accuracy']))
                env_stats["naive_bayes_f1_weighted"] = float(mean(scr['test_f1_macro']))
            except: #pragma: no cover
                pass

            try:
                #Average Node Learner [3,4]
                scr = cross_validate(RandomForestClassifier(n_estimators=50,criterion='entropy',max_depth=1), np_X, np_Y, scoring=('accuracy','f1_macro'))
                env_stats["average_node_accuracy"] = float(mean(scr['test_accuracy']))
                env_stats["average_node_f1_macro"] = float(mean(scr['test_f1_weighted']))
            except: #pragma: no cover
                pass

            try:
                #Best Node Learner [3,4]
                scr = cross_validate(DecisionTreeClassifier(criterion='entropy',max_depth=1), np_X, np_Y, scoring=('accuracy','f1_macro'))
                env_stats["best_node_accuracy"] = float(mean(scr['test_accuracy']))
                env_stats["best_node_f1_macro"] = float(mean(scr['test_f1_macro']))
            except: #pragma: no cover
                pass

            try:
                #pca effective dimensions [1]
                centered_x= np_X - np_X.mean(axis=0)
                pca_var = TruncatedSVD(n_components=min(np_X.shape[1]-1,1000)).fit(centered_x).explained_variance_ratio_
                env_stats["pca_top_1_pct"] = pca_var[0:1].sum()
                env_stats["pca_top_2_pct"] = pca_var[0:2].sum()
                env_stats["pca_top_3_pct"] = pca_var[0:3].sum()
                env_stats["pca_dims_95"  ] = int(sum(np.cumsum(pca_var)<.95)+1)
            except: #pragma: no cover
                pass

            #sklearn's CCA doesn't seem to work with sparse so I'm leaving it out for now depsite [3]

        except CobaExit:
            pass

        return env_stats

    def _entropy(self, items: Sequence[Hashable]) -> float:
        return -sum([count/len(items)*math.log2(count/len(items)) for count in collections.Counter(items).values()])

    def _entropy_normed(self, items: Sequence[Hashable]) -> float:
        return self._entropy(items)/(math.log2(len(set(items))) or 1)

    def _mutual_info(self, items1: Sequence[Hashable], items2: Sequence[Hashable]) -> float:
        return self._entropy(items1) + self._entropy(items2) - self._entropy(list(zip(items1,items2)))

    def _imbalance_ratio(self, items: list) -> float:
        #Equation (37) and (38) in [1]

        counts = collections.Counter(items).values()
        n      = len(items)
        k      = len(counts)

        if n == 0: return None
        if max(counts) == n: return 1

        IR = (k-1)/k*sum([c/(n-c) for c in counts])
        return 1 - 1/IR

    def _max_fisher_discriminant_ratio(self, X, Y) -> float:
        #Equation (3) in [1]
        #needs more testing

        feats    = list(range(len(X[0])))
        Y_set    = set(Y)
        Y_counts = collections.Counter(Y)

        X_by_f   = collections.defaultdict(list)
        X_by_fy  = collections.defaultdict(lambda: collections.defaultdict(list))

        for x,y in zip(X,Y):
            for f in feats:
                xf = x[f]
                X_by_fy[f][y].append(xf)
                X_by_f[f].append(xf)

        mean_f  = { f:      mean(X_by_f[f])                      for f in feats}
        mean_fy = { f: { y: mean(X_by_fy[f][y]) for y in Y_set } for f in feats}

        max_ratio = 0

        for f in feats:
            ratio_numer = sum([Y_counts[y]*(mean_fy[f][y]-mean_f[f])**2 for y in Y_set])
            ratio_denom = sum([(X_by_fy[f][y][i]-mean_fy[f][y])**2 for y in Y_set for i in range(Y_counts[y]) ])
            if ratio_denom != 0:
                max_ratio   = max(max_ratio,ratio_numer/ratio_denom)

        return 1/(1+max_ratio)

    def _max_directional_fisher_discriminant_ratio(self, X, Y) -> float:
        #equation (4) in [1]
        #this code is currently not used because it can take an incredibly
        #long time to calculate due to np.linalg.inv so we "no cover" it

        try:
            PackageChecker.sklearn('')

            import numpy as np
            from sklearn.covariance import shrunk_covariance

            Y_set = set(Y)
            X     = np.array(X).T #transpose so that equations align with paper
            Y     = np.array(Y)

            X_by_y = { y: X[:,Y==y]  for y in Y_set }
            OVO    = []

            for y1,y2 in combinations(Y_set,2):
                mu_c1 = X_by_y[y1].mean(axis=1).reshape(-1,1)
                mu_c2 = X_by_y[y2].mean(axis=1).reshape(-1,1)

                p1 = X_by_y[y1].shape[1]/(X_by_y[y1].shape[1]+X_by_y[y2].shape[1])
                p2 = X_by_y[y2].shape[1]/(X_by_y[y1].shape[1]+X_by_y[y2].shape[1])

                s1 = (X_by_y[y1]-mu_c1) @ (X_by_y[y1]-mu_c1).T
                s2 = (X_by_y[y2]-mu_c2) @ (X_by_y[y2]-mu_c2).T
                B  = (mu_c1-mu_c2) @ (mu_c1-mu_c2).T

                s1 = shrunk_covariance(s1)
                s2 = shrunk_covariance(s2)
                B  = shrunk_covariance(B)

                W = p1*s1+p2*s2
                d = np.linalg.inv(W) @ (mu_c1-mu_c2)

                OVO.append( ((d.T@B@d) / (d.T@W@d))[0,0] )

            return 1/(1+mean(OVO)) if OVO else None

        except np.linalg.LinAlgError:#pragma: no cover
            return None
        except CobaExit:#pragma: no cover
            return None

    def _volume_overlapping_region(self, X, Y) -> float:
        #equation (9) in [1]

        X_by_y = collections.defaultdict(list)
        for x,y in zip(X,Y): X_by_y[y].append(x)
        F_by_y = { y: list(zip(*x)) for y,x in X_by_y.items()}

        minmax = lambda f,y1,y2: min(max(F_by_y[y1][f]), max(F_by_y[y2][f]))
        maxmin = lambda f,y1,y2: max(min(F_by_y[y1][f]), min(F_by_y[y2][f]))
        maxmax = lambda f,y1,y2: max(max(F_by_y[y1][f]), max(F_by_y[y2][f]))
        minmin = lambda f,y1,y2: min(min(F_by_y[y1][f]), min(F_by_y[y2][f]))

        OVO = []

        for y1,y2 in combinations(set(Y),2):
            F2 = 1
            for f in range(len(X[0])):
                if maxmax(f,y1,y2)-minmin(f,y1,y2) != 0:
                    F2 *= max(0, (minmax(f,y1,y2)-maxmin(f,y1,y2)))/(maxmax(f,y1,y2)-minmin(f,y1,y2))

            OVO.append(F2)

        return mean(OVO) if OVO else None

    def _max_individual_feature_efficiency(self, X, Y) -> float:
        #equation (11) in [1]

        try:
            PackageChecker.numpy('ClassEnvironmentTask')
            import numpy as np

            X_by_f = np.array(list(zip(*X)))
            X_by_y = collections.defaultdict(list)
            for x,y in zip(X,Y): X_by_y[y].append(x)
            F_by_y  = { y: list(zip(*x)) for y,x in X_by_y.items()}
            L_by_fy = { f: { y:percentile(F_by_y[y][f],[0.0,1.0])  for y in F_by_y } for f in range(len(X[0])) }

            minmax = lambda f,y1,y2: min(L_by_fy[f][y1][1], L_by_fy[f][y2][1])
            maxmin = lambda f,y1,y2: max(L_by_fy[f][y1][0], L_by_fy[f][y2][0])

            OVO    = []

            for y1,y2 in combinations(set(Y),2):

                n_o = []
                for f in range(len(X[0])):
                    n_o.append(int(((X_by_f[f] <= minmax(f,y1,y2)) & (maxmin(f,y1,y2) <= X_by_f[f])).sum()))

                OVO.append(min(n_o)/len(X))

            return mean(OVO) if OVO else None

        except CobaExit:
            return None

    def _dense(self, X) -> Sequence[Sequence[float]]:

        if not isinstance(X[0],dict):
            return X
        else:
            #Convert the sparse dicts into a compact dense array.
            #This is required by for a number of the analysis in this task.
            feats = sorted(set().union(*X)) #sort to make unit tests repeatable
            dense_X = [ [0]*len(feats) for _ in range(len(X)) ]

            for i,x in enumerate(X):
                for j,f in enumerate(feats):
                        dense_X[i][j] = x.get(f,0)

            return dense_X

    def _bin(self, X: Sequence[Sequence[float]], n_bins:int, lower:float=0.05, upper:float=0.95) -> Sequence[Sequence[float]]:
        X_by_f = list(zip(*X))
        lim_f  = { f: percentile(X_by_f[f],[lower,upper]) for f in range(len(X_by_f)) }
        X_bin  = [ [ round((n_bins-1)*(x[f]-lim_f[f][0])/((lim_f[f][1]-lim_f[f][0]) or 1)) for f in range(len(X_by_f)) ] for x in X]
        return X_bin



import pickle

from pathlib import Path
from zipfile import ZipFile, ZIP_DEFLATED
from itertools import islice, repeat, chain
from typing import Union, Sequence, Mapping, Iterable, Any

from coba.context import CobaContext
from coba.primitives import Source, Sink, Filter, Environment, Interaction

class ObjectsToZipMember(Sink[Iterable[Sequence[object]]]):
    def __init__(self, zip:str):
        self._zip = zip
        self._start = 0
        if Path(self._zip).exists():
            for n in ZipFile(self._zip).namelist():
                try:
                    self._start = max(self._start,int(n)+1)
                except ValueError: #pragma: no cover
                    pass

    def write(self, envs: Iterable[Sequence[object]]) -> None:
        for i,env in enumerate(envs,self._start):
            with CobaContext.logger.time(f"Writing environment {i}..."):
                # Dump before opening the file so we
                # can keep it open for as short a time
                # as possible.
                dumps = list(map(pickle.dumps,env))
                with ZipFile(self._zip,mode='a',compression=ZIP_DEFLATED) as zip:
                    with zip.open(str(i),mode='w') as f:
                        f.writelines(dumps)

class ZipMemberToObjects(Source[Iterable[object]]):
    def __init__(self, zip:str, member:str):
        self._zip    = zip
        self._member = member

    def read(self) -> Iterable[object]:
        try:
            with ZipFile(self._zip) as z:
                with z.open(self._member) as m:
                    yield from map(pickle.load, repeat(m))
        except EOFError:
            pass

class EnvironmentsToObjects(Filter[Environment, Iterable]):
    def filter(self, env: Union[Environment,Sequence[Environment]]) -> Iterable[Iterable[object]]:
        with CobaContext.logger.time("Materializing environment..."):
            yield list(self._env_to_objects(env))

    def _env_to_objects(self,env):
        from coba import __version__ #imported here to avoid circular dependency
        yield {"version":2,"coba_version":__version__}
        yield env.params
        I = iter(env.read())
        batch = list(islice(I,1000))
        while batch:
            yield batch
            batch = list(islice(I,1000))

class EnvironmentFromObjects(Environment):
    def __init__(self, source: Source[Iterable[object]]) -> None:
        self._source = source

    @property
    def params(self) -> Mapping[str,Any]:
        return list(islice(self._source.read(),1,2))[0]

    def read(self) -> Iterable[Interaction]:

        I = iter(self._source.read())
        version_data = list(islice(I,2))[0]

        if version_data['version'] == 1: #pragma: no cover
            yield from I
        else:
            yield from chain.from_iterable(I)



from re import match
from sys import platform
from itertools import repeat, compress
from typing import Any, Dict, Union, Sequence, Mapping, Optional, Tuple, Literal

from coba.exceptions import CobaException
from coba.primitives import is_batch, Learner, Context, Action, Actions, Prob, Kwargs, Sparse, Namespaces
from coba.utilities import PackageChecker
from coba.learners.utilities import PMFInfoPredictor

Feature       = Union[str,int,float]
Features      = Union[Feature, Sequence[Feature], Dict[str,Union[int,float]]]
VW_Features   = Sequence[Union[str,int,Tuple[Union[str,int],Union[int,float]],Dict[str,Union[int,float]]]]
VW_Namespaces = Dict[str,VW_Features]

DEFAULT_NAMESPACE_INTERACTIONS = (1, 'a', 'ax', 'axx')

def make_args(vw_kwargs: Dict[str, Any], namespace_interactions: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS) -> Sequence[str]:
    """
    Turn settings into VW command line arguments.

        Args:
        vw_kwargs: Keyword argument dict that gets translated into VW CLI arguments
        namespace_interactions: A sequence of namespace interactions to use during learning.
    """
    args = []
    vw_kwargs['quiet'] = vw_kwargs.get('quiet', True)

    for k,v in vw_kwargs.items():
        if v is not False and v is not None:
            if not k.startswith("-"):
                k = ("-" if len(k)==1 else "--") + k
            args.append(k if (v is True) else f"{k} {v}")

    noconstant = sum([n for n in namespace_interactions if isinstance(n, (int, float))]) == 0
    ignore_linear = set(['x', 'a']) - set(namespace_interactions)
    interactions = [n for n in namespace_interactions if isinstance(n, str) and len(n) > 1]

    if noconstant:
        args.append(f"--noconstant")

    for interaction in interactions:
        args.append(f"--interactions {interaction}")

    for ignore in ignore_linear:
        args.append(f"--ignore_linear {ignore}")

    return args

class VowpalMediator:
    """A class to handle all communication between Coba and VW."""

    def __init__(self) -> None:
        self._vw = None
        self._args = ""
        self._ns_keys = {}
        self._ns_keys_cnt = 0

        PackageChecker.vowpalwabbit('VowpalMediator')

    @property
    def is_initialized(self) -> bool:
        """Indicate whether init_learner has been called previously."""
        return self._vw is not None

    @property
    def params(self) -> dict:
        return {}

    def init_learner(self, args:str, label_type: int) -> 'VowpalMediator':
        """Create a VW learner from a command line arg string.

        Args:
            args: The command line arg string to use for VW learner creation.
            label_type: The label type this VW learner will take.
                - 1 : `simple`__
                - 2 : `multiclass`__
                - 3 : `cost sensitive`__
                - 4 : `contextual bandit`__
                - 5 : max (deprecated)
                - 6 : `conditional contextual bandit`__
                - 7 : `slates`__
                - 8 : `continuous actions`__

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Input-format#simple
        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Input-format#multiclass
        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Input-format#cost-sensitive
        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Input-format#contextual-bandit
        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Conditional-Contextual-Bandit#vw-text-format
        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Slates#text-format
        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/CATS,-CATS-pdf-for-Continuous-Actions#vw-text-format
        """
        from vowpalwabbit import pyvw, __version__

        if self._vw is not None:
            raise CobaException("We cannot initilaize a VW learner twice in a single mediator.")

        self._version = [ int(v) for v in __version__.split(".") ]
        self._vw = pyvw.Workspace(args) if self._version[0] >= 9 else pyvw.vw(args)
        self._label_type = pyvw.LabelType(label_type) if self._version[0] >= 9 else label_type
        self._example_init = pyvw.Example if self._version[0] >= 9 else pyvw.example
        self._args = args

        return self

    def predict(self, example: Any) -> Any:
        """Predict for an example created by the mediator."""
        pred = self._vw.predict(example)
        self._vw.finish_example(example)
        return pred

    def learn(self, example: Any) -> None:
        """Learn for an example created by the mediator."""
        self._vw.learn(example)
        self._vw.finish_example(example)

    def finish(self):
        if self.is_initialized:
            self._vw.finish()

    def make_example(self, namespaces: Namespaces, label:Optional[str] = None) -> Any:
        """Create a VW example.

        Args:
            ns: The features grouped by namespace in this example.
            label: An optional label (required if this example will be used for learning).
        """
        ns = dict(self._prep_namespaces(namespaces))

        ex = self._example_init(self._vw, None, self._label_type)
        ex.push_feature_dict(self._vw, ns)
        if label is not None: ex.set_label_string(label)
        ex.setup_example()

        return ex

    def make_examples(self, shared: Namespaces, uniques: Sequence[Namespaces], labels:Optional[Sequence[str]] = None) -> Sequence[Any]:
        """Create a list of VW examples.

        Args:
            shared: The features grouped by namespace in this example.
            separates: The features, grouped by namespace, unique to each each example.
            label: An optional label (required if this example will be used for learning).
        """

        labels     = repeat(None) if labels is None else labels
        vw_shared  = dict(self._prep_namespaces(shared))
        vw_uniques = list(map(dict,map(self._prep_namespaces,uniques)))

        examples = []
        for vw_unique, label in zip(vw_uniques,labels):
            ex = self._example_init(self._vw, None, self._label_type)
            ex.push_feature_dict(self._vw, {**vw_shared, **vw_unique})
            if label: ex.set_label_string(label)
            ex.setup_example()
            examples.append(ex)

        return examples

    def _prep_namespaces(self, namespaces: Namespaces) -> VW_Namespaces:
        """Turn a collection of coba formatted namespaces into VW format."""

        if self._version < [9,2] or (self._version == [9,2,0] and "linux" in platform): #pragma: no cover
            #the strange type checks below were faster than traditional methods when performance testing
            for ns, feats in namespaces.items():
                if not feats and feats != 0 and feats != "":
                    continue
                elif feats.__class__ is str:
                    yield (ns, [f"{self._get_ns_keys(ns,0)}={feats}"])
                elif feats.__class__ is int or feats.__class__ is float:
                    yield (ns, [(self._get_ns_keys(ns,0), feats)])
                else:
                    feats = feats.items() if isinstance(feats, Sparse) else zip(self._get_ns_keys(ns,len(feats)),feats)
                    yield (ns, [f"{k}={v}" if v.__class__ is str else (k, v) for k,v in feats if v!= 0])
        else: #pragma: no cover
            #the strange type checks below were faster than traditional methods when performance testing
            for ns, feats in namespaces.items():
                if not feats and feats != 0 and feats != "":
                    continue
                elif feats.__class__ is str:
                    yield (ns, {f"{self._get_ns_keys(ns,0)}={feats}":1})
                elif feats.__class__ is int or feats.__class__ is float:
                    yield (ns, {self._get_ns_keys(ns,0): feats})
                elif isinstance(feats,Sparse):
                    no_str = not any(map(isinstance,feats.values(),repeat(str))) # most-performant (still 2x slower than no check)
                    #no_str = str not in set(map(attrgetter("__class__"),feats.values()))
                    #no_str = not any(map(is_,map(attrgetter("__class__"),feats.values()),repeat(str)))
                    #no_str = True
                    if no_str:
                        yield (ns, feats.copy())
                    else:
                        new = feats.copy()
                        for k,v in feats.items():
                            if isinstance(v,str):
                                new[f"{k}={v}"] = 1
                                del new[k]
                        yield (ns,new)
                else:
                    no_str = not any(map(isinstance,compress(feats,feats),repeat(str)))

                    K = self._get_ns_keys(ns,len(feats))
                    V = feats

                    if no_str:
                        d = dict(compress(zip(K,V),V))
                    else:
                        d={}
                        for k,v in compress(zip(K,V),V):
                            if v.__class__ is str:
                                d[f"{k}={v}"] = 1
                            else:
                                d[k] = v

                    yield (ns, d)

    def _get_ns_keys(self, ns:str, length: int) -> Union[str,Sequence[str]]:

        if ns not in self._ns_keys:
            if not length:
                self._ns_keys[ns] = str(self._ns_keys_cnt)
                self._ns_keys_cnt += 1
            else:
                self._ns_keys[ns] = list(map(str,range(self._ns_keys_cnt,self._ns_keys_cnt+length)))
                self._ns_keys_cnt += length

        return self._ns_keys[ns]

    def __str__(self) -> str:
        return self._args

class VowpalLearner(Learner):
    """A Vowpal Wabbit wrapper.

    Remarks:
        This learner requires that the Vowpal Wabbit package be installed. This package can be
        installed via `pip install vowpalwabbit`. To learn more about solving contextual bandit
        problems with Vowpal Wabbit see `here`__ and `here`__.

    __ https://vowpalwabbit.org/tutorials/contextual_bandits.html
    __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
    """

    def __init__(self, args: str = "--cb_explore_adf --epsilon 0.05 --interactions ax --interactions axx --ignore_linear x --random_seed 1 --quiet", vw: VowpalMediator = None, seed: int = 1) -> None:
        """Instantiate a VowpalLearner.

        Args:
            args: Command line arguments to instantiate a Vowpal Wabbit contextual bandit learner. For
                examples and documentation on how to instantiate VW learners from command line arguments
                see `here`__. We require that either cb, cb_adf, cb_explore, or cb_explore_adf is used.
                When we format examples for VW context features are placed in the 'x' namespace and action
                features, when relevant, are placed in the 'a' namespace.
            vw: A mediator able to communicate with VW. This should not need to ever be changed.
            seed: A seed for a random number generation.
        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
        """

        if "--cb" not in args:
            raise CobaException("VowpalLearner was instantiated without a cb flag. One of the cb flags must be defined.")

        self._args    = args
        self._explore = "--cb_explore" in args
        self._adf     = "--cb_adf"     in args or "--cb_explore_adf" in args

        self._n_actions = None
        self._actions   = None

        if not self._adf:
            n_action_str = match("--cb.*?\s*(\d*)\\b.*", args).group(1)
            self._n_actions = int(n_action_str) if n_action_str else None
            #useful for removing the n_actions from args
            #re.sub("(--cb)\s*(\d+)", '\g<1>', "--cb 123", count=1)

        self._vw = vw or VowpalMediator()
        self._pred = PMFInfoPredictor(self._pmf,seed)

    @property
    def params(self) -> Mapping[str, Any]:
        return {"family": "vw", 'args': self._args.replace("--quiet","").strip(), **self._vw.params, 'seed': self._pred.seed}

    def _dflt_context_ns(self,context) -> Namespaces:
        return {'x':context} if not isinstance(context,Namespaces) else context

    def _dflt_actions_ns(self,actions) -> Optional[Sequence[Namespaces]]:
        if self._adf:
            is_ns = actions and isinstance(actions[0],Namespaces)
            return actions if is_ns else [{'a':a} for a in actions]
        return None

    def _pmf(self, context, actions):
        if not self._vw.is_initialized and is_batch(context): #pragma: no cover
            raise CobaException("VW learner does not support batched calls.")
        if not self._vw.is_initialized and self._adf:
            self._vw.init_learner(self._args, 4)
        if not self._vw.is_initialized and not self._adf and self._n_actions:
            self._vw.init_learner(self._args, 4)
        if not self._vw.is_initialized and not self._adf and not self._n_actions:
            self._n_actions = len(actions)
            args            = self._args.replace('--cb_explore','').replace('--cb','')
            args            = f"--cb_explore {len(actions)} " if self._explore else f"--cb {len(actions)} " + args
            args            = args.strip()
            self._vw.init_learner(args,4)
        if not self._adf and not self._actions:
            self._actions = actions
        if not self._adf and actions != self._actions:
            raise CobaException("Actions are only allowed to change between predictions when using `adf`.")
        if not self._adf and len(actions) != self._n_actions:
            raise CobaException("The number of actions doesn't match the `--cb` action count given in args.")

        context = self._dflt_context_ns(context)
        adfs    = self._dflt_actions_ns(actions)

        if self._adf and self._explore:
            probs = self._vw.predict(self._vw.make_examples(context, adfs, None))
            if probs is None: probs = [1/len(actions)] * len(actions)

        if self._adf and not self._explore:
            losses    = self._vw.predict(self._vw.make_examples(context,adfs, None))
            min_loss  = min(losses)
            min_bools = [s == min_loss for s in losses]
            min_count = sum(min_bools)
            probs     = [ int(min_indicator)/min_count for min_indicator in min_bools ]

        if not self._adf and self._explore:
            probs = self._vw.predict(self._vw.make_example(context, None))
            if probs is None: probs = [1/len(actions)] * len(actions)

        if not self._adf and not self._explore:
            index = self._vw.predict(self._vw.make_example(context, None))
            probs = [ int(i==index) for i in range(1,len(actions)+1) ]

        return probs, {'actions':actions}

    def score(self, context: 'Context', actions: 'Actions', action: 'Action') -> 'Prob':
        return self._pred.score(context,actions,action)

    def predict(self, context: 'Context', actions: 'Actions') -> Tuple['Action','Prob','Kwargs']:
        return self._pred.predict(context,actions)

    def learn(self, context: 'Context', action: 'Action', reward: float, probability: float, actions: 'Actions' = None) -> None:
        if not self._vw.is_initialized and self._adf:
            self._vw.init_learner(self._args, 4)
        if not self._vw.is_initialized and not self._adf and self._n_actions:
            self._vw.init_learner(self._args, 4)
        if not self._vw.is_initialized and not self._adf and not self._n_actions:
            raise CobaException("When using `cb` without `adf` predict must be called before learn to initialize the vw learner")

        if actions is None: actions = [action]

        index   = actions.index(action)
        labels  = self._labels(actions, index, reward, probability)
        label   = labels[index]

        context = self._dflt_context_ns(context)
        adfs    = self._dflt_actions_ns(actions)

        if self._adf:
            self._vw.learn(self._vw.make_examples(context, adfs, labels))
        else:
            self._vw.learn(self._vw.make_example(context, label))

    def _labels(self,actions,index,reward:float,prob:float) -> Sequence[Optional[str]]:
        labels = [None]*len(actions)
        labels[index] = f"{index+1}:{round(-reward,5)}:{round(prob,5)}"
        return labels

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.finish()

    def finish(self):
        """Finish all pending work (e.g., write buffers to disk)."""
        try:
            self._vw.finish()
        except AttributeError:
            #we do these checks because when mocking up for unittests
            #it is possible that _vw actually isn't ever initialized
            pass

    def __del__(self):
        self.finish()

class VowpalEpsilonLearner(VowpalLearner):
    """Epsilon-greedy exploration with a VW contextual bandit learner.

    More information on VW exploration algorithms can be found `here`__.

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
    """

    def __init__(self,
                 epsilon: float = 0.05,
                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
                 seed: Optional[int] = 1,
                 **kwargs) -> None:
        """Instantiate a VowpalEpsilonLearner.

        Args:
            epsilon: The probability that we will explore instead of exploit.
            features: A list of namespaces and interactions to use when learning reward functions.
            seed: The seed used by VW to generate any necessary random numbers.
            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
        """
        self._params = {'epsilon': epsilon, 'features': features}

        vw_kwargs = {
            "cb_explore_adf": True,
            "epsilon": epsilon,
            "random_seed": seed
        }
        vw = kwargs.pop('vw', None)
        vw_kwargs.update(kwargs)
        vw_args_string = " ".join(make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))

        super().__init__(vw_args_string, vw)

    @property
    def params(self) -> Mapping[str,Any]:
        return {**self._params, **super().params}

class VowpalSoftmaxLearner(VowpalLearner):
    """Softmax exploration with a VW contextual bandit learner.

    More information on VW exploration algorithms can be found `here`__.

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
    """

    def __init__(self,
                 softmax: float=10,
                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
                 seed: Optional[int] = 1,
                 **kwargs) -> None:
        """Instantiate a VowpalSoftmaxLearner.

        Args:
            softmax: An exploration parameter with 0 indicating predictions should be completely random
                and infinity indicating that predictions should be greedy. For more information see `lambda`__.
            features: A list of namespaces and interactions to use when learning reward functions.
            seed: The seed used by VW to generate any necessary randomness.
            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
        """
        self._params = {'softmax': softmax, 'features': features}

        vw_kwargs = {
            "cb_explore_adf": True,
            "softmax": True,
            "lambda": softmax,
            "random_seed": seed
        }
        vw = kwargs.pop('vw', None)
        vw_kwargs.update(kwargs)
        vw_args_string = " ".join(make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
        super().__init__(vw_args_string, vw)

    @property
    def params(self) -> Mapping[str,Any]:
        return {**self._params, **super().params}

class VowpalBagLearner(VowpalLearner):
    """Bootstrap aggregated policy exploration with a VW contextual bandit learner.

    More information on VW exploration algorithms can be found `here`__.

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
    """

    def __init__(self,
                 bag: int = 5,
                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
                 seed: Optional[int] = 1,
                 **kwargs) -> None:
        """Instantiate a VowpalBagLearner.

        Args:
            bag: This value determines the number of policies which will be learned and must be greater
                than 0. Each policy is trained using bootstrap aggregation, making each policy unique. During
                prediction a random policy will be selected according to a uniform distribution and followed.
            features: A list of namespaces and interactions to use when learning reward functions.
            seed: The seed used by VW to generate any necessary random numbers.
            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
        """
        self._params = {'bag': bag, 'features': features}

        vw_kwargs = {
            "cb_explore_adf": True,
            "bag": bag,
            "random_seed": seed
        }
        vw = kwargs.pop('vw', None)
        vw_kwargs.update(kwargs)
        vw_args_string = " ".join(make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
        super().__init__(vw_args_string, vw)

    @property
    def params(self) -> Mapping[str,Any]:
        return {**self._params, **super().params}

class VowpalCoverLearner(VowpalLearner):
    """Online Cover exploration with a VW contextual bandit learner.

    For more information on this algorithm see Agarwal et al. (2014) and `here`__.

    References:
        Agarwal, Alekh, Daniel Hsu, Satyen Kale, John Langford, Lihong Li, and Robert Schapire. "Taming
        the monster: A fast and simple algorithm for contextual bandits." In International Conference on
        Machine Learning, pp. 1638-1646. 2014.

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
    """

    def __init__(self,
                 cover: int = 5,
                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
                 seed: Optional[int] = 1,
                 **kwargs) -> None:
        """Instantiate a VowpalCoverLearner.

        Args:
            cover: The number of policies which will be learned (must be greater than 0).
            features: A list of namespaces and interactions to use when learning reward functions.
            seed: The seed used by VW to generate any necessary random numbers.
            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
        """
        self._params = {'cover': cover, 'features': features}

        vw_kwargs = {
            "cb_explore_adf": True,
            "cover": cover,
            "random_seed": seed
        }
        vw = kwargs.pop('vw', None)
        vw_kwargs.update(kwargs)
        vw_args_string = " ".join(make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
        super().__init__(vw_args_string, vw)

    @property
    def params(self) -> Mapping[str,Any]:
        return {**self._params, **super().params}

class VowpalRndLearner(VowpalLearner):
    """RND exploration with a VW contextual bandit learner.

    Inspired by Random Network Distillation, this explorer constructs an auxiliary prediction
    problem whose expected target value is zero and uses the prediction magnitude to construct
    a confidence interval. In the contextual bandit case this is equivalent to a randomized
    approximation to the LinUCB bound.

    For more information see the `wiki`__.

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
    """

    def __init__(self,
                 rnd: int = 3,
                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
                 epsilon: Optional[float] = 0.025,
                 rnd_alpha: Optional[float] = None,
                 rnd_invlambda: Optional[float] = None,
                 seed: Optional[int] = 1,
                 **kwargs) -> None:
        """Instantiate a VowpalRndLearner.

        Args:
            rnd: Number of predictors
            features: A list of namespaces and interactions  to use when learning reward functions
            epsilon: Uniform exploration term for stabilization
            rnd_alpha: Increase for more exploration on a repeated example
            rnd_invlambda: Increase for more exploration on examples with new features/actions
            seed: The seed used by VW to generate any necessary random numbers
            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
        """
        self._params = {
            'rnd': rnd,
            'features': features,
            'epsilon': epsilon,
            'rnd_alpha': rnd_alpha,
            'rnd_invlambda': rnd_invlambda
        }

        vw_kwargs = {
            "cb_explore_adf": True,
            "rnd": rnd,
            "epsilon": epsilon,
            "rnd_alpha": rnd_alpha,
            "rnd_invlambda": rnd_invlambda,
            "random_seed": seed
        }
        vw = kwargs.pop('vw', None)
        vw_kwargs.update(kwargs)
        vw_args_string = " ".join(make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
        super().__init__(vw_args_string, vw)

    @property
    def params(self) -> Mapping[str,Any]:
        return {**self._params, **super().params}

class VowpalRegcbLearner(VowpalLearner):
    """RegCB exploration with a VW contextual bandit learner.

    For more information on this algorithm see Foster et al. (2014) and `here`__.

    References:
        Foster, D., Agarwal, A., Dudik, M., Luo, H. & Schapire, R.. (2018). Practical Contextual
        Bandits with Regression Oracles. Proceedings of the 35th International Conference on Machine
        Learning, in Proceedings of Machine Learning Research 80:1539-1548.

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
    """

    def __init__(self,
                 mode: Literal["optimistic","elimination"] = "elimination",
                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
                 seed: Optional[int] = 1,
                 **kwargs) -> None:
        """Instantiate a VowpalRegcbLearner.

        Args:
            mode: Indicate that exploration should only predict the optimal upper bound action or
                should use an elimination technique to remove actions that no longer seem plausible
                and pick randomly from the remaining actions.
            features: A list of namespaces and interactions to use when learning reward functions.
            seed: The seed used by VW to generate any necessary random numbers.
            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
        """
        self._params = {'mode': mode, 'features': features}

        vw_kwargs = {
            "cb_explore_adf": True,
        }
        if mode == "elimination":
            vw_kwargs["regcb"] = True
        else:
            vw_kwargs["regcbopt"] = True
        vw_kwargs["random_seed"] = seed

        vw = kwargs.pop('vw', None)
        vw_kwargs.update(kwargs)
        vw_args_string = " ".join(make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
        super().__init__(vw_args_string, vw)

    @property
    def params(self) -> Mapping[str,Any]:
        return {**self._params, **super().params}

class VowpalSquarecbLearner(VowpalLearner):
    """SquareCB exploration with a VW contextual bandit learner.

    For more information on this algorithm see Foster et al. (2020) and `here`__.

    References:
        Foster, D.& Rakhlin, A.. (2020). Beyond UCB: Optimal and Efficient Contextual Bandits with Regression
        Oracles. Proceedings of the 37th International Conference on Machine Learning, in Proceedings of Machine
        Learning Research 119:3199-3210.

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
    """

    def __init__(self,
                 mode: Literal["standard","elimination"] = "standard",
                 gamma_scale: float = 10,
                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
                 seed: Optional[int] = 1,
                 **kwargs) -> None:
        """Instantiate a VowpalSquarecbLearner.

        Args:
            mode: Indicates iwhether all actions should be considered for exploration on each step or actions
                which no longer seem plausible should be eliminated.
            gamma_scale: Controls how quickly squarecb exploration converges to a greedy policy. The larger the
                gamma_scale the faster the algorithm will converge to a greedy policy. This value is the same
                as gamma in the original paper.
            features: A list of namespaces and interactions to use when learning reward functions.
            seed: The seed used by VW to generate any necessary random numbers.
            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
        """
        self._params = {'mode': mode, 'gamma_scale': gamma_scale, 'features': features}

        vw_kwargs = {
            "cb_explore_adf": True,
            "squarecb": True,
            "gamma_scale": gamma_scale,
            "random_seed": seed
        }
        if mode == "elimination":
            vw_kwargs["elim"] = True
        vw = kwargs.pop('vw', None)
        vw_kwargs.update(kwargs)
        vw_args_string = " ".join(make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
        super().__init__(vw_args_string, vw)

    @property
    def params(self) -> Mapping[str,Any]:
        return {**self._params, **super().params}

class VowpalOffPolicyLearner(VowpalLearner):
    """No exploration with a VW contextual bandit learner.

    This wrapper performs policy learning without any exploration. This is
    correct when training examples come from a logging policy that controls
    all exploration.

    Remarks:
        More information on VW exploration algorithms can be found `here`__.

        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
    """

    def __init__(self,
                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
                 seed: Optional[int] = 1,
                 **kwargs) -> None:
        """Instantiate a VowpalOffPolicyLearner.

        Args:
            features: A list of namespaces and interactions to use when learning reward functions.
            seed: The seed used by VW to generate any necessary random numbers.
            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
        """
        self._params = {'features': features}

        vw_kwargs = {
            "cb_adf": True,
            "random_seed": seed
        }
        vw = kwargs.pop('vw', None)
        vw_kwargs.update(kwargs)
        vw_args_string = " ".join(make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
        super().__init__(vw_args_string, vw)

    @property
    def params(self) -> Mapping[str,Any]:
        return {**self._params, **super().params}



import json
import collections.abc

from urllib import request
from pathlib import Path
from zipfile import ZipFile, BadZipFile
from itertools import compress
from typing import Sequence, overload, Union, Iterable, Iterator, Optional, Tuple, Callable, Mapping, Type, Literal, Any

import coba.json
from coba                 import pipes
from coba.context         import CobaContext, DiskCacher, DecoratedLogger, ExceptLog, NameLog, StampLog
from coba.primitives      import Context, Action, Source, Learner, Environment, EnvironmentFilter
from coba.random          import CobaRandom
from coba.primitives      import Dense, Sparse
from coba.pipes           import Pipes, HttpSource, IterableSource, DataFrameSource, DiskSource, NextSource, IdentitySource
from coba.exceptions      import CobaException
from coba.multiprocessing import CobaMultiprocessor
from coba.results         import Result, Missing

from coba.environments.templates  import EnvironmentsTemplateV1, EnvironmentsTemplateV2
from coba.environments.openml     import OpenmlSimulation
from coba.environments.synthetics import LinearSyntheticSimulation, NeighborsSyntheticSimulation, BanditSyntheticSimulation
from coba.environments.synthetics import KernelSyntheticSimulation, MLPSyntheticSimulation, LambdaSimulation
from coba.environments.supervised import SupervisedSimulation
from coba.environments.results    import ResultEnvironment

from coba.environments.filters   import Repr, Batch, Chunk, Logged, Finalize
from coba.environments.filters   import Binary, Shuffle, Take, Sparsify, Densify, Reservoir, Cycle, Scale, Unbatch
from coba.environments.filters   import Slice, Impute, Where, Riffle, Sort, Flatten, Cache, Params, Grounded
from coba.environments.filters   import OpeRewards, Noise, BatchSafe

from coba.environments.serialized import EnvironmentFromObjects, EnvironmentsToObjects, ZipMemberToObjects, ObjectsToZipMember

class Environments(collections.abc.Sequence, Sequence[Environment]):
    """A friendly API for common environment functionality."""

    @staticmethod
    def cache_dir(path:Union[str,Path]='~/.cache/coba') -> Type['Environments']:
        """Set the cache directory for openml sources.

        Args:
            path: A path to a directory to cache openml sources.

        Returns:
            The Environments class.
        """
        CobaContext.cacher = DiskCacher(path)
        return Environments

    @staticmethod
    def from_template(source: Union[str,Source[Iterable[str]]], **user_vars) -> 'Environments':
        """Create Environments from a template file.

        Args:
            **user_vars: overrideable template variables

        Returns:
            An Environments object.
        """
        try:
            return Environments(*EnvironmentsTemplateV2(source, **user_vars).read())
        except Exception as e: #pragma: no cover
            try:
                #try previous version of definition files. If previous versions also fail
                #then we raise the exception given by the most up-to-date version so that
                #changes can be made to conform to the latest version.
                return Environments(*EnvironmentsTemplateV1(source).read())
            except:
                raise e

    @staticmethod
    def from_prebuilt(name:str) -> 'Environments':
        """Create Environments from a pre-built definition.

        Args:
            name: The desired pre-built environment.

        Returns:
            An Environments object.
        """

        repo_url       = "https://github.com/mrucker/coba_prebuilds/blob/main"
        definition_url = f"{repo_url}/{name}/index.json?raw=True"

        try:
            definition_txt = HttpSource(definition_url).read()
        except request.HTTPError as e:
            if e.code != 404:
                raise
            else:
                root_dir_text = HttpSource("https://api.github.com/repos/mrucker/coba_prebuilds/contents/").read()
                root_dir_json = json.loads(root_dir_text)
                known_names   = [ obj['name'] for obj in root_dir_json if obj['name'] != "README.md" ]
                raise CobaException(f"The given prebuilt name, {name}, couldn't be found. Known names are: {known_names}")

        definition_txt = definition_txt.replace('"./', f'"{repo_url}/{name}/')
        definition_txt = definition_txt.replace('.json"', '.json?raw=True"')
        return Environments.from_template(IterableSource([definition_txt]))

    @staticmethod
    def from_bandit_synthetic(
        n_interactions: Optional[int],
        n_actions: int = 5,
        seed: Union[int,Sequence[int]] =1) -> 'Environments':
        """Create Environments from bandit environment.

        Args:
            n_interactions: An optional integer indicating the number of interactions in the simulation.
            n_actions: The number of actions each interaction should have.
            seed: An integer seed to guarantee repeatability.

        Returns:
            An Environments object.
        """

        seed = [seed] if not isinstance(seed,collections.abc.Sequence) else seed
        args = (n_interactions, n_actions)

        return Environments([BanditSyntheticSimulation(*args, s) for s in seed])

    @staticmethod
    def from_linear_synthetic(
        n_interactions: int,
        n_actions:int = 5,
        n_context_features: int = 5,
        n_action_features: int = 5,
        n_coefficients:Optional[int] = 5,
        reward_features: Sequence[str] = ["a","xa"],
        seed: Union[int,Sequence[int]] = 1) -> 'Environments':
        """Create Environments from linear reward functions.

        Args:
            n_interactions: The number of interactions the simulation should have.
            n_actions: The number of actions each interaction should have.
            n_context_features: The number of features each context should have.
            n_action_features: The number of features each action should have.
            n_coefficients The number of non-zero weights in the final reward function.
            reward_features: The features in the simulation's linear reward function.
            seed: The seed used to generate all random values. If seed is a list then
                an separate environment will be created for each seed.

        Returns:
            An Environments object.
        """

        seed = [seed] if not isinstance(seed,collections.abc.Sequence) else seed
        args = (n_interactions, n_actions, n_context_features, n_action_features, n_coefficients, reward_features)

        return Environments([LinearSyntheticSimulation(*args, s) for s in seed])

    @staticmethod
    def from_neighbors_synthetic(
        n_interactions: int,
        n_actions: int = 5,
        n_context_features: int = 5,
        n_action_features: int = 5,
        n_neighborhoods: int = 30,
        seed: Union[int,Sequence[int]] = 1) -> 'Environments':
        """Create Environments from nearest neighbors reward functions.

        Args:
            n_interactions: The number of interactions the simulation should have.
            n_actions: The number of actions each interaction should have.
            n_context_features: The number of features each context should have.
            n_action_features: The number of features each action should have.
            n_neighborhoods: The number of distinct reward value neighborhoods.
            seed: The seed used to generate all random values. If seed is a list then
                an separate environment will be created for each seed.

        Returns:
            An Environments object.
        """

        seed = [seed] if not isinstance(seed,collections.abc.Sequence) else seed
        args = (n_interactions, n_actions, n_context_features, n_action_features, n_neighborhoods)

        return Environments([NeighborsSyntheticSimulation(*args, s) for s in seed])

    @staticmethod
    def from_kernel_synthetic(
        n_interactions:int,
        n_actions:int = 5,
        n_context_features:int = 5,
        n_action_features:int = 5,
        n_exemplars:int = 5,
        kernel: Literal['linear','polynomial','exponential','gaussian'] = 'gaussian',
        degree: int = 3,
        gamma: float = 1,
        seed: Union[int,Sequence[int]] = 1) -> 'Environments':
        """Create Environments from kernel-based reward functions.

        Args:
            n_interactions: The number of interactions the simulation should have.
            n_actions: The number of actions each interaction should have.
            n_context_features: The number of features each context should have.
            n_action_features: The number of features each action should have.
            n_exemplars: The number of exemplar context-action pairs.
            kernel: The family of the kernel basis functions.
            degree: This argument is only relevant when using polynomial kernels.
            gamma: This argument is only relevant when using exponential kernels.
            seed: The seed used to generate all random values. If seed is a list then
                an separate environment will be created for each seed.

        Returns:
            An Environments object.
        """

        seed = [seed] if not isinstance(seed,collections.abc.Sequence) else seed
        args = (n_interactions, n_actions, n_context_features, n_action_features, n_exemplars, kernel, degree, gamma)

        return Environments([KernelSyntheticSimulation(*args, s) for s in seed])

    @staticmethod
    def from_mlp_synthetic(
        n_interactions:int,
        n_actions:int = 5,
        n_context_features:int = 5,
        n_action_features:int = 5,
        seed: Union[int,Sequence[int]] = 1) -> 'Environments':
        """Create Environments from kernel-based reward functions.

        Args:
            n_interactions: The number of interactions the simulation should have.
            n_actions: The number of actions each interaction should have.
            n_context_features: The number of features each context should have.
            n_action_features: The number of features each action should have.
            seed: The seed used to generate all random values. If seed is a list then
                an separate environment will be created for each seed.

        Remarks:
            The MLP architecture has a single hidden layer with sigmoid
            activation and one output value calculated from a random
            linear combination of the hidden layer's output.

        Returns:
            An Environments object.
        """

        seed = [seed] if not isinstance(seed,collections.abc.Sequence) else seed
        args = (n_interactions, n_actions, n_context_features, n_action_features)

        return Environments([MLPSyntheticSimulation(*args, s) for s in seed])

    @overload
    @staticmethod
    def from_openml(
        data_id: Union[int,Sequence[int]],
        drop_missing: bool = True,
        take: int = None,
        *,
        target:str = None,
        label_type:Literal['c','r','m'] = None) -> 'Environments':
        ...
        """Create Environments from openml datasets.

        Args:
            data_id: The data id for a dataset on openml (i.e., openml.org/d/{id}).
                If data_id is a list then an environment will be created for each id.
            drop_missing: Drop data rows with missing values.
            take: The interaction count for the simulation (selected at random).
            target: The column that should be marked as the label in the source.
            label_type: Is the label a classification, regression or multilabel type.

        Returns:
            An Environments object.
        """

    @overload
    @staticmethod
    def from_openml(
        *,
        task_id: Union[int,Sequence[int]],
        drop_missing: bool = True,
        take: int = None,
        target:str = None,
        label_type:Literal['m','c','r'] = None) -> 'Environments':
        ...
        """Create Environments from openml datasets.

        Args:
            task_id: The task id for a task on openml (i.e., openml.org/t/{id}).
                If task_id is a list then an environment will be created for each id.
            drop_missing: Drop data rows with missing values.
            take: The interaction count for the simulation (selected at random).
            target: The column that should be marked as the label in the source.
            label_type: Is the label a classification, regression or multilabel type.

        Returns:
            An Environments object.
        """

    @staticmethod
    def from_openml(*args,**kwargs) -> 'Environments':
        """Create Environments from openml datasets.

        Args:
            data_id: The data id for a dataset on openml (i.e., openml.org/d/{id}).
                If data_id is a list then an environment will be created for each id.
            task_id: The task id for a task on openml (i.e., openml.org/t/{id}).
                If task_id is a list then an environment will be created for each id.
            drop_missing: Drop data rows with missing values.
            take: The interaction count for the simulation (selected at random).
            target: The column that should be marked as the label in the source.
            label_type: Is the label a classification, regression or multilabel type.

        Returns:
            An Environments object.

        """

        kwargs.update(zip(['data_id','drop_missing','take'], args))

        if 'data_id' in kwargs and isinstance(kwargs['data_id'],int):
            kwargs['data_id'] = [kwargs['data_id']]

        if 'task_id' in kwargs and isinstance(kwargs['task_id'],int):
            kwargs['task_id'] = [kwargs['task_id']]

        if 'data_id' in kwargs:
            return Environments(*[OpenmlSimulation(data_id=id, **kwargs) for id in kwargs.pop('data_id')])
        else:
            return Environments(*[OpenmlSimulation(task_id=id, **kwargs) for id in kwargs.pop('task_id')])

    @overload
    @staticmethod
    def from_supervised(
        source: Source[Union[Iterable[Dense], Iterable[Sparse],Iterable[Tuple[Any,Any]]]],
        label_col: Union[int,str] = None,
        label_type: Literal["c","r","m"] = None,
        take: int = None) -> 'Environments':
        """Create Environments from supervised datasets.

        Args:
            source: A source that reads the supervised data. Coba natively
                provides support for csv, arff, libsvm, and manik data sources.
            label_col: The header or index of the label in each example. If None
                the source must return an iterable of tuple pairs where the first
                item are features and the second item is a label.
            label_type: Is the label a classification, regression or multilabel type.
                If None the label type will be inferred based on the data source.
            take: The interaction count for the simulation (selected at random).

        Returns:
            An Environments object.
        """
        ...

    @overload
    @staticmethod
    def from_supervised(
        X = Sequence[Any],
        Y = Sequence[Any],
        label_type: Literal["c","r","m"] = None) -> 'Environments':
        """Create Environments using a supervised dataset.

        Args:
            X: The features to use when creating contexts.
            Y: The labels to use when creating actions and rewards.
            label_type: Is the label a classification, regression or multilabel type.
                If None the label type will be inferred based on the data source.

        Returns:
            An Environments object.
        """
        ...

    @staticmethod
    def from_supervised(*args, **kwargs) -> 'Environments':
        """Create Environments using a supervised dataset.

        Args:
            source: A source that reads the supervised data. Coba natively
                provides support for csv, arff, libsvm, and manik data sources.
            label_col: The header or index of the label in each example. If None
                the source must return an iterable of tuple pairs where the first
                item are features and the second item is a label.
            X: The features to use when creating contexts.
            Y: The labels to use when creating actions and rewards.
            label_type: Is the label a classification, regression or multilabel type.
                If None the label type will be inferred based on the data source.
            take: The interaction count for the simulation (selected at random).

        Returns:
            An Environments object.
        """
        return Environments(SupervisedSimulation(*args, **kwargs))

    @staticmethod
    def from_save(path:str) -> 'Environments':
        """Create Environments from an Environments save file.

        Args:
            path: Path to an Environments save file.

        Returns:
            An Environments object.
        """
        envs = []
        for name in ZipFile(path).namelist():
            envs.append(EnvironmentFromObjects(ZipMemberToObjects(path,name)))
        return Environments(envs)

    @overload
    def from_lambda(self,
        n_interactions: Optional[int],
        context       : Callable[[int               ],Context         ],
        actions       : Callable[[int,Context       ],Sequence[Action]],
        reward        : Callable[[int,Context,Action],float           ]) -> 'Environments':
        """Create a SimulatedEnvironment from lambda functions."""

    @overload
    def from_lambda(self,
        n_interactions: Optional[int],
        context       : Callable[[int               ,CobaRandom],Context         ],
        actions       : Callable[[int,Context       ,CobaRandom],Sequence[Action]],
        reward        : Callable[[int,Context,Action,CobaRandom],float           ],
        seed          : int) -> 'Environments':
        """Create a SimulatedEnvironment from lambda functions."""

    @staticmethod
    def from_lambda(*args,**kwargs) -> 'Environments':
        """Create a SimulatedEnvironment from lambda functions."""
        return Environments(LambdaSimulation(*args,**kwargs))

    @staticmethod
    def from_dataframe(dataframe) -> 'Environments':
        """Create Environments from a dataframe.

        Args:
            dataframe: Create Environments from this. There will
                only be one environment whose interactions will
                be the rows of the dataframe.

        Returns:
            An Environments object.
        """
        return Environments(DataFrameSource(dataframe))

    @staticmethod
    def from_result(result:Union[str,Result]) -> 'Environments':
        """Create Environments from a given Result file.

        Args:
            result: The path to results or the results of an experiment. One
                environment will be created for every environment in the
                Experiment that produced the results.

        Remarks:
            We assume that 'context', 'action', probability', and 'reward' was
            recorded during the experiment.

        Returns:
            An Environments object.
        """

        if isinstance(result,str):
            class InteractionSource:
                def __init__(self,source:Source[str]):
                    self._source = source
                def read(self) -> Mapping[str,Sequence]:
                    return coba.json.loads(self._source.read())[2]['_packed']

            env_rows = collections.defaultdict(dict)
            lrn_rows = collections.defaultdict(dict)
            val_rows = collections.defaultdict(dict)
            interactions = []

            for (loc,line) in DiskSource(result,include_loc=True).read():
                if line.strip():
                    trx = json.loads(line)
                    if trx[0] == "E": env_rows[trx[1]].update(trx[2])
                    if trx[0] == "L": lrn_rows[trx[1]].update(trx[2])
                    if trx[0] == "V": val_rows[trx[1]].update(trx[2])
                    if trx[0] == "I": interactions.append((loc,*trx[1],0)[:4] )

            envs = []
            for loc, env_id, lrn_id, val_id in interactions:
                env_params = env_rows.get(env_id)
                lrn_params = lrn_rows.get(lrn_id)
                val_params = val_rows.get(val_id)
                int_source = InteractionSource(NextSource(DiskSource(result,start_loc=loc)))

                envs.append(ResultEnvironment(int_source,env_params,lrn_params,val_params))
        else:
            env_rows = { d.pop('environment_id'): d for d in result.environments.to_dicts() }
            lrn_rows = { d.pop('learner_id'): d for d in result.learners.to_dicts() }
            val_rows = { d.pop('evaluator_id'): d for d in result.evaluators.to_dicts() }

            all_hdrs = result.interactions.columns[4:]

            envs = []
            for (env_id,lrn_id,val_id), cols in result.interactions.groupby(3,all_hdrs):
                env_params = env_rows.get(env_id)
                lrn_params = lrn_rows.get(lrn_id)
                val_params = val_rows.get(val_id)

                keep = [ not all(c == Missing for c in col) for col in cols]
                hdrs = list(compress(all_hdrs,keep))
                cols = list(compress(cols,keep))

                int_source = IdentitySource(dict(zip(hdrs,cols)))
                envs.append(ResultEnvironment(int_source,env_params,lrn_params,val_params))


        return Environments(envs)

    @staticmethod
    def from_custom(*environments: Union[Environment, Sequence[Environment]]):
        """Create Environments from Environment.

        Args:
            *environments: Create an Environments from the environments.

        Returns:
            An Environments object.
        """
        return Environments(*environments)

    @staticmethod
    def from_feurer(drop_missing: bool = True) -> 'Environments':
        """Create Environments from the Feurer benchmark.

        Args:
            drop_missing: Exclude interactions with missing context features.

        Remarks:
            The description of the benchmark is provided at https://arxiv.org/abs/2007.04074.
            For Task ids 232, 3044, 75105, and 211723 every row has a missing feature. These
            environments will be empty when drop_missing is True. Task id 189866 has been
            updated to 361282, a new version of the original dataset that fixes api issues
            with the old dataset.

        Returns:
            An Environments object.
        """

        task_ids = [232,236,241,245,253,254,256,258,260,262,267,271,273,275,279,288,336,340,
                    2119,2120,2121,2122,2123,2125,2356,3044,3047,3048,3049,3053,3054,3055,
                    75089,75092,75093,75097,75098,75100,75105,75108,75109,75112,75114,75115,
                    75116,75118,75120,75121,75125,75126,75127,75129,75131,75133,75134,75136,
                    75139,75141,75142,75143,75146,75147,75148,75149,75153,75154,75156,75157,
                    75159,75161,75163,75166,75169,75171,75173,75174,75176,75178,75179,75180,
                    75184,75185,75187,75192,75193,75195,75196,75199,75210,75212,75213,75215,
                    75217,75219,75221,75223,75225,75232,75233,75234,75235,75236,75237,75239,
                    75250,126021,126024,126025,126026,126028,126029,126030,126031,146574,146575,
                    146576,146577,146578,146583,146586,146592,146593,146594,146596,146597,146600,
                    146601,146602,146603,146679,166859,166866,166872,166875,166882,166897,166905,
                    166906,166913,166915,166931,166932,166944,166950,166951,166953,166956,166957,
                    166958,166959,166970,166996,167083,167085,167086,167087,167088,167089,167090,
                    167094,167096,167097,167099,167100,167101,167103,167104,167105,167106,167149,
                    167152,167161,167168,167181,167184,167185,167190,167200,167201,167202,167203,
                    167204,167205,168785,168791,168792,168793,168794,168795,168796,168797,168798,
                    189779,189786,189828,189829,189836,189840,189841,189843,189844,189845,189846,
                    189858,189859,189860,189861,189862,189863,189864,189865,361282,189869,189870,
                    189871,189872,189873,189874,189875,189878,189880,189881,189882,189883,189884,
                    189887,189890,189893,189894,189899,189900,189902,189905,189906,189908,189909,
                    190154,190155,190156,190157,190158,190159,211720,211721,211722,211723,211724]

        return Environments.from_openml(task_id=task_ids,drop_missing=drop_missing)

    def __init__(self, *environments: Union[Environment, Sequence[Environment]]):
        """Instantiate an Environments class.

        Args:
            *environments: Base environments.
        """
        is_collection = lambda env: isinstance(env, (collections.abc.Sequence,collections.abc.Generator))
        self._envs = []
        for env in environments:
            if is_collection(env):
                self._envs.extend(env)
            else:
                self._envs.append(env)
        self._envs = list(map(Pipes.join,self._envs))

    def binary(self) -> 'Environments':
        """Transform reward values to 1 or 0.

        Remarks:
            Reward values are either 1 (if max reward) or 0 (not max reward).

        Returns:
            An Environments object.
        """
        return self.filter(Binary())

    def sparse(self, context:bool = True, action:bool = False) -> 'Environments':
        """Ensure that features are sparse.

        Args:
            context: Sparsify context features.
            action: Sparsify action features.

        Returns:
            An Environments object.
        """
        return self.filter(Sparsify(context,action))

    def dense(self, n_feats:int, method:Literal['lookup','hashing'], context:bool = True, action:bool = False) -> 'Environments':
        """Ensure that features are dense.

        Args:
            n_feats: The number of features densified environment should have.
            method: How sparse features are turned into dense features. The hashing
                trick is more memory efficient but may have collisions. The lookup
                method is less memory efficient but guaranteed to have no collisions.
            context: Densify context features.
            action: Densify action features.

        Returns:
            An Environments object.
        """


        make_dense = lambda: Densify(n_feats=n_feats, method=method, context=context, action=action)
        return Environments([Pipes.join(env,make_dense()) for env in self._envs])

    @overload
    def shuffle(self, seed: int = 1) -> 'Environments':
        ...

    @overload
    def shuffle(self, seeds: Iterable[int]) -> 'Environments':
        ...

    @overload
    def shuffle(self, *, n:int) -> 'Environments':
        ...

    def shuffle(self, *args,**kwargs) -> 'Environments':
        """Shuffle interaction order.

        Args:
            seed: The seed determining shuffle order.
            seeds: Sequence of seeds determining shuffle order.
                A new environment is made for every seed where
                the only difference is the order of interactions.
            n: The number of shuffling orders to produce. Equivalent
                to `shuffle(seeds=range(n))`.

        Returns:
            An Environments object.
        """

        flat = lambda a: next(pipes.Flatten().filter([a]))

        if kwargs and 'n' in kwargs:
            seeds = range(kwargs['n'])
        else:
            seeds = flat(kwargs.get('seed',kwargs.get('seeds',args)))

        if seeds != 0 and not seeds: seeds = [1]
        if isinstance(seeds,int): seeds = [seeds]

        shuffled = self.filter([Shuffle(seed) for seed in seeds])

        #Experience has shown that most of the time we want to sort.
        #This doesn't change the experiment results. It simply makes it
        #easier monitor an experiment while it runs in the background.
        ordered = sorted(shuffled, key=lambda env: env.params.get('shuffle_seed',0))

        return Environments(ordered)

    def sort(self, *keys: Union[str,int,Sequence[Union[str,int]]]) -> 'Environments':
        """Sort interactions by features.

        Args:
            *keys: The index or keys for context features.

        Returns:
            An Environments object.
        """
        return self.filter(Sort(*keys))

    def riffle(self, spacing: int, seed: int = 1) -> 'Environments':
        """Riffle shuffle interactions.

        Args:
            spacing: The number of interactions from the beginning
                between each interaction shuffled in from the end.
            seed: The seed used to determine the location of each
                ending interaction when placed within its beginning
                space.

        Returns:
            An Environments object.
        """
        return self.filter(Riffle(spacing, seed))

    def cycle(self, after: int) -> 'Environments':
        """Cycle reward values.

        Useful for testing a learner's response to a non-stationary noise.

        Args:
            after: The number of interactions to wait before cycling reward values.

        Returns:
            An Environments object.
        """
        if isinstance(after,(int,float)): after = [after]
        return self.filter([Cycle(a) for a in after])

    def params(self, params: Mapping[str,Any]) -> 'Environments':
        """Add params to environments.

        Args:
            params: Parameter values to add to each Environment in Environments.

        Returns:
            An Environments object.
        """
        return self.filter(Params(params))

    def take(self, n_interactions: int, strict: bool = False) -> 'Environments':
        """Take the first n interactions.

        Args:
            n_interactions: The maximum number of interactions to take.
            strict: Do not take any interactions if there are not at least n_interactions.

        Returns:
            An Environments object.
        """
        return self.filter(Take(n_interactions, strict))

    def slice(self, start: Optional[int], stop: Optional[int]=None, step:int = 1) -> 'Environments':
        """Take a slice of interactions.

        Args:
            start: The starting index for the slice.
            stop: The finishing index for the slice (exclusive).
            step: The step size between each item in the slice.

        Returns:
            An Environments object.
        """
        return self.filter(Slice(start,stop,step))

    def reservoir(self, n_interactions: int, seeds: Union[int,Sequence[int]]=1, strict:bool = False) -> 'Environments':
        """Take n random interactions.

        Args:
            n_interactions: The maximum number of interactions to sample.
            seeds: The seed to use to randomly generate the sample. If a
                sequence of seeds is provided then multiple samples are drawn.
            strict: Do not draw a sample if there are not at least n_interactions.

        Returns:
            An Environments object.
        """
        if isinstance(seeds,int): seeds = [seeds]
        return self.filter([Reservoir(n_interactions,strict=strict,seed=seed) for seed in seeds])

    def scale(self,
        shift: Union[float,Literal["min","mean","med"]] = "min",
        scale: Union[float,Literal["minmax","std","iqr","maxabs"]] = "minmax",
        targets:Literal["context"] = "context",
        using: Optional[int] = None) -> 'Environments':
        """Scale and shift features.

        Args:
            shift: The statistic to use to shift each context feature.
            scale: The statistic to use to scale each context feature.
            target: The target data we wish to scale in the environment.
            using: The number of interactions to use when calculating statistics.

        Remarks:
            For example, `scale('mean', 'std')` would standardize all context features
            while `scale('med', 'iqr')` would apply what sklearn calls a RobustScaler
            to all context features.

        Returns:
            An Environments object.
        """
        if isinstance(targets,str): targets = [targets]
        for t in targets: self = self.filter(Scale(shift, scale, t, using))
        return self

    def impute(self,
        stats: Union[Literal["mean","median","mode"],Sequence[Literal["mean","median","mode"]]] = "mean",
        indicator:bool = True,
        using: Optional[int] = None) -> 'Environments':
        """Impute missing data.

        Args:
            stats: The statistic to use for imputation. If a sequence is provided they
                will be applied in order. This is useful if different statistics should
                be applied for different types.
            indicator: Indicates whether a binary feature should be added for missingness.
            using: The number of interactions to use to calculate imputation statistics.

        Returns:
            An Environments object.
        """
        if isinstance(stats,str): stats = [stats]
        envs = self
        for stat in stats:
            envs = self.filter(Impute(stat, indicator, using))
        return envs

    def where(self,*,
            n_interactions: Union[int,Tuple[Optional[int],Optional[int]]] = None,
            n_actions: Union[int,Tuple[Optional[int],Optional[int]]] = None,
            n_features: Union[int,Tuple[Optional[int],Optional[int]]] = None) -> 'Environments':
        """Select for characteristics.

        Args:
            n_interactions: The min, max or exact number of interactions an Environment must have.
            n_actions: The min, max or exact number of actions an interaction must have.
            n_features: The minimum, maximum or exact number of features interactions must have.

        Returns:
            An Environments object.
        """
        return self.filter(Where(n_interactions=n_interactions,n_actions=n_actions,n_features=n_features))

    def noise(self,
        context: Union[Tuple[float,float],Tuple[str,float,float],Callable[[float,CobaRandom], float]] = None,
        action : Union[Tuple[float,float],Tuple[str,float,float],Callable[[float,CobaRandom], float]] = None,
        reward : Union[Tuple[float,float],Tuple[str,float,float],Callable[[float,CobaRandom], float]] = None,
        seed   : Union[int,Sequence[int]] = 1) -> 'Environments':
        """Add noise to values.

        Args:
            context: Shape parameters for a distribution or a callable that returns a noisy value.
            action : Shape parameters for a distribution or a callable that returns a noisy value.
            reward : Shape parameters for a distribution or a callable that returns a noisy value.
            seed: The seed for all random values. If a sequence of seeds is given then multiple
                environments will be created with each using distinct noise filters from the seeds.

        Remarks:
            Supported distributions defined via tuple are:
                * random gaussian: (`mean`,`std`)
                * random integer : ('i',`inclusive min`,`inclusive max`)
                * random gaussian: ('g',`mean`,`std`)

        Returns:
            An Environments object.
        """
        if isinstance(seed,(int,float)): seed = [seed]
        return self.filter([Noise(context,action,reward,s) for s in seed])

    def flatten(self) -> 'Environments':
        """Flatten contexts and actions.

        Examples:
            An interaction {'context': [[1,2],3]} would become {'context':[1,2,3]}.

        Returns:
            An Environments object.
        """
        return self.filter(Flatten())

    def materialize(self) -> 'Environments':
        """Materialize and cache all environments.

        Remarks:
            Ideal for stateful environments such as Jupyter Notebook where
            environments can be saved in memory and re-used between experiments.

        Returns:
            An Environments object.
        """
        #we use pipes.cache directly because the environment cache will copy
        #which we don't need when we materialize an environment in memory

        nocache = lambda p: not isinstance(p,pipes.Cache) or p.protected
        materialized = []
        for env in map(self._finalize,self._envs):
            if not isinstance(env[-1],pipes.Cache):
                env = Pipes.join(*filter(nocache,env), pipes.Cache(None,True))
                list(env.read()) #force read to pre-load cache
            materialized.append(env)

        return Environments(materialized)

    def grounded(self, n_users: int, n_normal:int, n_words:int, n_good:int, seed:int=1) -> 'Environments':
        """Transform simulated interactions to IGL interactions.

        Args:
            n_users: The number of users in the grounded environment.
            n_normal: The number of users with normal grounded behavior.
            n_words: The number of potential feedback words for users.
            n_good: The number of words that mean good out of the n_words.
            seed: Seed for all random values.

        Remarks:
            See `here`__ for more on interaction grounded learning.

        Returns:
            An Environments object.

        __ https://proceedings.mlr.press/v139/xie21e.html
        """
        return self.filter(Grounded(n_users, n_normal, n_words, n_good, seed))

    def repr(self,
        cat_context:Literal["onehot","onehot_tuple","string"] = "onehot",
        cat_actions:Literal["onehot","onehot_tuple","string"] = "onehot") -> 'Environments':
        """Change representation of categorical data.

        Args:
            cat_context: How to represent categorical data in contexts.
            cat_actions: How to represent categorical data in actions.

        Returns:
            An Environments object.
        """
        return self.filter(Repr(cat_context,cat_actions))

    def batch(self, batch_size:int, batch_type: Literal['list','torch'] = 'list') -> 'Environments':
        """Batch interactions.

        Args:
            batch_size: The number of interactions in a batched interaction.
            batch_type: The type of batch for interaction values.

        Returns:
            An Environments object.
        """
        return self.filter(Batch(batch_size, batch_type))

    def chunk(self, cache:bool = True) -> 'Environments':
        """Add a chunk pipe.

        Args:
            cache: output before the chunk should be cached.

        Remarks:
            This is useful if an early part of an Environments pipeline
            takes considerable time to evaluate (e.g., `logged`). Placing
            a chunk after the long running filter means that the filter
            will only be executed one time and the results will be resued.

        Returns:
            An Environments object.
        """
        envs = Environments([Pipes.join(env,Chunk()) for env in self._envs])
        return envs.cache() if cache else envs

    def logged(self, learners: Union[Learner,Sequence[Learner]], seed:Optional[float] = 1.23) -> 'Environments':
        """Transform simulated interactions to logged interactions.

        Args:
            learners: The learners that will be used as the logging policy.
                An environment will be created for every learner provided.
            seed: The seed for used for all random number generation.

        Remarks:
            Adds 'action', 'reward', and 'probability' to interactions with 'context',
            'actions', and 'rewards'.

        Returns:
            An Environments object.
        """
        if not isinstance(learners, collections.abc.Sequence): learners = [learners]
        return self.filter([Logged(learner, seed) for learner in learners ])

    def unbatch(self):
        """Unbatch interactions.

        Remarks:
            The `unbatch` command is the inverse of `batch`.

        Returns:
            An Environments object.
        """

        return self.filter(Unbatch())

    def ope_rewards(self, rewards_type:Literal['IPS','DM','DR'] = None):
        """Transform logged interactions to simulated interactions.

        Args:
            rewards_type: How to estimate the rewards function from the logged
            data (i.e., inverse propensity score, direct method, or doubly robust).

        Remarks:
            Adds 'rewards' to interactions with 'context', 'action', 'reward',
            and 'probability'.

        Returns:
            An Environments object.
        """
        if isinstance(rewards_type,str): rewards_type = [rewards_type]
        return self.filter([OpeRewards(r) for r in rewards_type])

    def save(self, path: str, processes:int=1, overwrite:bool=False) -> 'Environments':
        """Save Environments to disk.

        Args:
            path: The location to save Environments (the file will be a zip archive).
            processes: The number of process to use when generating environments.
            overwrite: Indicate if an existing file at Path should be overwritten.

        Returns:
            An Environments object.
        """

        self_envs = list(self)
        if Path(path).exists():
            try:
                path_envs   = Environments.from_save(path)
                path_params = [e.params for e in path_envs]
                self_params = [e.params for e in self_envs]

                try:
                    while path_params:
                        param_index_in_self = self_params.index(path_params.pop())
                        self_params.pop(param_index_in_self)
                        self_envs.pop(param_index_in_self)
                except ValueError:
                    #there is a param in the file that isn't in self
                    is_equal = False
                else:
                    is_equal = True

                if is_equal and not self_envs:
                    return path_envs
                if not is_equal and overwrite:
                    Path(path).unlink()
                if not is_equal and not overwrite:
                    raise CobaException("The Environments save file does not match the actual Environments and overwite is False.")

            except BadZipFile:
                if overwrite:
                    Path(path).unlink()
                else:
                    raise CobaException("The given save file appears to be corrupted. Please check it and delete if it is unusable.")

        objs  = CobaMultiprocessor(EnvironmentsToObjects(),processes)
        disk  = ObjectsToZipMember(path)
        decor = [StampLog()] if processes ==1 else [NameLog(), StampLog()]

        CobaContext.logger = DecoratedLogger([ExceptLog()], CobaContext.logger, decor)
        disk.write(objs.filter(self_envs))
        CobaContext.logger = CobaContext.logger.undecorate()

        return Environments.from_save(path)

    def cache(self) -> 'Environments':
        """Add a caching pipe.

        Remarks:
            Results from earlier steps in the pipeline
            will re-used during later pipeline steps.

        Returns:
            An Environments object.
        """
        return Environments([Pipes.join(env,Cache(25)) for env in self._envs])

    def filter(self, filter: Union[EnvironmentFilter,Sequence[EnvironmentFilter]]) -> 'Environments':
        """Apply custom filter to Environments.

        Args:
            filter: The filters to apply to self. If a list of filters is
                provided then a new pipeline is created for each filter.

        Returns:
            An Environments object.
        """
        filters = filter if isinstance(filter, collections.abc.Sequence) else [filter]
        return Environments([Pipes.join(env,f) for env in self._envs for f in filters])

    def _finalize(self, env: Environment) -> Environment:
        is_finalize = lambda e: isinstance(e,BatchSafe) and isinstance(e._filter,Finalize)
        return env if any(map(is_finalize,env)) else Pipes.join(env,BatchSafe(Finalize()))

    def __getitem__(self, index:Any) -> Union[Environment,'Environments']:
        if isinstance(index,slice):
            return Environments(self._envs[index])
        else:
            return self._finalize(self._envs[index])

    def __iter__(self) -> Iterator[Environment]:
        return iter(map(self._finalize,self._envs))

    def __len__(self) -> int:
        return len(self._envs)

    def __add__(self, other: 'Environments') -> 'Environments':
        return Environments(self._envs+other._envs)

    def __str__(self) -> str:
        return "\n".join([f"{i+1}. {e}" for i,e in enumerate(self)])

    def _ipython_display_(self):
        #pretty print in jupyter notebook (https://ipython.readthedocs.io/en/stable/config/integrating.html)
        print(str(self))
